@@ -48,13 +48,12 @@
 
 #include <assert.h>
 #include <dirent.h>
-#include <dlfcn.h>
 #include <errno.h>
-#include <pthread.h>
+#include <limits.h>
 #include <stdarg.h>
 #include <stddef.h>
-#include <stdlib.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 
@@ -96,6 +95,7 @@
 
 typedef struct _Module {
 	CK_FUNCTION_LIST_PTR funcs;
+	CK_C_INITIALIZE_ARGS init_args;
 	int ref_count;
 
 	/* Registered modules */
@@ -103,21 +103,14 @@ typedef struct _Module {
 	hashmap *config;
 
 	/* Loaded modules */
-	void *dl_module;
+	dl_module_t dl_module;
 
-	/* Initialized modules */
-	CK_C_INITIALIZE_ARGS init_args;
-	int initialize_count;
+	/* Initialization, mutex must be held */
+	mutex_t initialize_mutex;
+	int initialize_called;
+	thread_t initialize_thread;
 } Module;
 
-/*
- * This is the mutex that protects the global data of this library
- * and the pkcs11 proxy module. Note that we *never* call into our
- * underlying pkcs11 modules while holding this mutex. Therefore it
- * doesn't have to be recursive and we can keep things simple.
- */
-pthread_mutex_t _p11_mutex = PTHREAD_MUTEX_INITIALIZER;
-
 /*
  * Shared data between threads, protected by the mutex, a structure so
  * we can audit thread safety easier.
@@ -134,20 +127,15 @@ static struct _Shared {
 static CK_RV
 create_mutex (CK_VOID_PTR_PTR mut)
 {
-	pthread_mutex_t *pmutex;
-	int err;
+	mutex_t *pmutex;
 
 	if (mut == NULL)
 		return CKR_ARGUMENTS_BAD;
 
-	pmutex = malloc (sizeof (pthread_mutex_t));
+	pmutex = malloc (sizeof (mutex_t));
 	if (!pmutex)
 		return CKR_HOST_MEMORY;
-	err = pthread_mutex_init (pmutex, NULL);
-	if (err == ENOMEM)
-		return CKR_HOST_MEMORY;
-	else if (err != 0)
-		return CKR_GENERAL_ERROR;
+	_p11_mutex_init (pmutex);
 	*mut = pmutex;
 	return CKR_OK;
 }
@@ -155,17 +143,12 @@ create_mutex (CK_VOID_PTR_PTR mut)
 static CK_RV
 destroy_mutex (CK_VOID_PTR mut)
 {
-	pthread_mutex_t *pmutex = mut;
-	int err;
+	mutex_t *pmutex = mut;
 
 	if (mut == NULL)
 		return CKR_MUTEX_BAD;
 
-	err = pthread_mutex_destroy (pmutex);
-	if (err == EINVAL)
-		return CKR_MUTEX_BAD;
-	else if (err != 0)
-		return CKR_GENERAL_ERROR;
+	_p11_mutex_uninit (pmutex);
 	free (pmutex);
 	return CKR_OK;
 }
@@ -173,36 +156,24 @@ destroy_mutex (CK_VOID_PTR mut)
 static CK_RV
 lock_mutex (CK_VOID_PTR mut)
 {
-	pthread_mutex_t *pmutex = mut;
-	int err;
+	mutex_t *pmutex = mut;
 
 	if (mut == NULL)
 		return CKR_MUTEX_BAD;
 
-	err = pthread_mutex_lock (pmutex);
-	if (err == EINVAL)
-		return CKR_MUTEX_BAD;
-	else if (err != 0)
-		return CKR_GENERAL_ERROR;
+	_p11_mutex_lock (pmutex);
 	return CKR_OK;
 }
 
 static CK_RV
 unlock_mutex (CK_VOID_PTR mut)
 {
-	pthread_mutex_t *pmutex = mut;
-	int err;
+	mutex_t *pmutex = mut;
 
 	if (mut == NULL)
 		return CKR_MUTEX_BAD;
 
-	err = pthread_mutex_unlock (pmutex);
-	if (err == EINVAL)
-		return CKR_MUTEX_BAD;
-	else if (err == EPERM)
-		return CKR_MUTEX_NOT_LOCKED;
-	else if (err != 0)
-		return CKR_GENERAL_ERROR;
+	_p11_mutex_unlock (pmutex);
 	return CKR_OK;
 }
 
@@ -214,14 +185,17 @@ free_module_unlocked (void *data)
 	assert (mod);
 
 	/* Module must be finalized */
-	assert (mod->initialize_count == 0);
+	assert (mod->initialize_called == 0);
+	assert (mod->initialize_thread == 0);
 
 	/* Module must have no outstanding references */
 	assert (mod->ref_count == 0);
 
 	if (mod->dl_module)
-		dlclose (mod->dl_module);
-	hash_free (mod->config);
+		_p11_module_close (mod->dl_module);
+
+	_p11_mutex_uninit (&mod->initialize_mutex);
+	_p11_hash_free (mod->config);
 	free (mod->name);
 	free (mod);
 }
@@ -240,6 +214,7 @@ alloc_module_unlocked (void)
 	mod->init_args.LockMutex = lock_mutex;
 	mod->init_args.UnlockMutex = unlock_mutex;
 	mod->init_args.flags = CKF_OS_LOCKING_OK;
+	_p11_mutex_init (&mod->initialize_mutex);
 
 	return mod;
 }
@@ -262,9 +237,14 @@ build_path (const char *dir, const char *filename)
 	assert (filename);
 
 	len = snprintf (NULL, 0, "%s/%s", dir, filename) + 1;
-	if (len <= 0 || len > PATH_MAX)
+	if (len <= 0)
 		return NULL;
 
+#ifdef PATH_MAX
+	if (len > PATH_MAX)
+		return NULL;
+#endif
+
 	if (!(path = malloc (len)))
 		return NULL;
 
@@ -282,16 +262,16 @@ dlopen_and_get_function_list (Module *mod, const char *path)
 	assert (mod);
 	assert (path);
 
-	mod->dl_module = dlopen (path, RTLD_LOCAL | RTLD_NOW);
+	mod->dl_module = _p11_module_open (path);
 	if (mod->dl_module == NULL) {
-		_p11_message ("couldn't load module: %s: %s", path, dlerror ());
+		_p11_message ("couldn't load module: %s: %s", path, _p11_module_error ());
 		return CKR_GENERAL_ERROR;
 	}
 
-	gfl = dlsym (mod->dl_module, "C_GetFunctionList");
+	gfl = _p11_module_symbol (mod->dl_module, "C_GetFunctionList");
 	if (!gfl) {
 		_p11_message ("couldn't find C_GetFunctionList entry point in module: %s: %s",
-		              path, dlerror ());
+		              path, _p11_module_error ());
 		return CKR_GENERAL_ERROR;
 	}
 
@@ -302,7 +282,7 @@ dlopen_and_get_function_list (Module *mod, const char *path)
 		return rv;
 	}
 
-	debug ("opened module: %s", path);
+	_p11_debug ("opened module: %s", path);
 	return CKR_OK;
 }
 
@@ -324,14 +304,14 @@ load_module_from_file_unlocked (const char *path, Module **result)
 	}
 
 	/* Do we have a previous one like this, if so ignore load */
-	prev = hash_get (gl.modules, mod->funcs);
+	prev = _p11_hash_get (gl.modules, mod->funcs);
 
 	if (prev != NULL) {
-		debug ("duplicate module %s, using previous", path);
+		_p11_debug ("duplicate module %s, using previous", path);
 		free_module_unlocked (mod);
 		mod = prev;
 
-	} else if (!hash_set (gl.modules, mod->funcs, mod)) {
+	} else if (!_p11_hash_set (gl.modules, mod->funcs, mod)) {
 		free_module_unlocked (mod);
 		return CKR_HOST_MEMORY;
 	}
@@ -347,7 +327,7 @@ expand_module_path (const char *filename)
 	char *path;
 
 	if (is_relative_path (filename)) {
-		debug ("module path is relative, loading from: %s", P11_MODULE_PATH);
+		_p11_debug ("module path is relative, loading from: %s", P11_MODULE_PATH);
 		path = build_path (P11_MODULE_PATH, filename);
 	} else {
 		path = strdup (filename);
@@ -369,9 +349,9 @@ take_config_and_load_module_unlocked (char **name, hashmap **config)
 	assert (config);
 	assert (*config);
 
-	module_filename = hash_get (*config, "module");
+	module_filename = _p11_hash_get (*config, "module");
 	if (module_filename == NULL) {
-		debug ("no module path for module, skipping: %s", *name);
+		_p11_debug ("no module path for module, skipping: %s", *name);
 		return CKR_OK;
 	}
 
@@ -380,7 +360,7 @@ take_config_and_load_module_unlocked (char **name, hashmap **config)
 		return CKR_HOST_MEMORY;
 
 	/* The hash map will take ownership of the variable */
-	if (!hash_set (*config, "module", path)) {
+	if (!_p11_hash_set (*config, "module", path)) {
 		free (path);
 		return CKR_HOST_MEMORY;
 	}
@@ -401,35 +381,35 @@ take_config_and_load_module_unlocked (char **name, hashmap **config)
 		return rv;
 	}
 
-	prev = hash_get (gl.modules, mod->funcs);
+	prev = _p11_hash_get (gl.modules, mod->funcs);
 
-	/* Replace previous module that was loaded explicitly? */
-	if (prev && !prev->name) {
-		mod->ref_count = prev->ref_count;
-		mod->initialize_count = prev->initialize_count;
-		prev->ref_count = 0;
-		prev->initialize_count = 0;
-		prev = NULL; /* freed by hash_set below */
-	}
+	/* If same module was loaded previously, just take over config */
+	if (prev && !prev->name && !prev->config) {
+		prev->name = mod->name;
+		mod->name = NULL;
+		prev->config = mod->config;
+		mod->config = NULL;
+		free_module_unlocked (mod);
 
-	/* Refuse to load duplicate module */
-	if (prev) {
+	/* Ignore duplicate module */
+	} else if (prev) {
 		_p11_message ("duplicate configured module: %s: %s", mod->name, path);
 		free_module_unlocked (mod);
-		return CKR_OK;
-	}
 
+	/* Add this new module to our hash table */
+	} else {
+		if (!_p11_hash_set (gl.modules, mod->funcs, mod)) {
+			free_module_unlocked (mod);
+			return CKR_HOST_MEMORY;
+		}
+
+	}
 	/*
 	 * We support setting of CK_C_INITIALIZE_ARGS.pReserved from
 	 * 'x-init-reserved' setting in the config. This only works with specific
 	 * PKCS#11 modules, and is non-standard use of that field.
 	 */
-	mod->init_args.pReserved = hash_get (mod->config, "x-init-reserved");
-
-	if (!hash_set (gl.modules, mod->funcs, mod)) {
-		free_module_unlocked (mod);
-		return CKR_HOST_MEMORY;
-	}
+	mod->init_args.pReserved = _p11_hash_get (mod->config, "x-init-reserved");
 
 	return CKR_OK;
 }
@@ -460,7 +440,7 @@ load_registered_modules_unlocked (void)
 	                                  P11_USER_CONFIG_MODULES);
 	if (configs == NULL) {
 		rv = (errno == ENOMEM) ? CKR_HOST_MEMORY : CKR_GENERAL_ERROR;
-		hash_free (config);
+		_p11_hash_free (config);
 		return rv;
 	}
 
@@ -471,13 +451,13 @@ load_registered_modules_unlocked (void)
 	 * Now go through each config and turn it into a module. As we iterate
 	 * we steal the values of the config.
 	 */
-	hash_iterate (configs, &iter);
-	while (hash_next (&iter, &key, NULL)) {
-		if (!hash_steal (configs, key, (void**)&name, (void**)&config))
+	_p11_hash_iterate (configs, &iter);
+	while (_p11_hash_next (&iter, &key, NULL)) {
+		if (!_p11_hash_steal (configs, key, (void**)&name, (void**)&config))
 			assert (0 && "not reached");
 
 		/* Is this a critical module, should abort loading of others? */
-		critical = _p11_conf_parse_boolean (hash_get (config, "critical"), 0);
+		critical = _p11_conf_parse_boolean (_p11_hash_get (config, "critical"), 0);
 
 		rv = take_config_and_load_module_unlocked (&name, &config);
 
@@ -486,16 +466,16 @@ load_registered_modules_unlocked (void)
 		 * by the above function call.
 		 */
 		free (name);
-		hash_free (config);
+		_p11_hash_free (config);
 
 		if (critical && rv != CKR_OK) {
 			_p11_message ("aborting initializationg because module '%s' was marked as critical");
-			hash_free (configs);
+			_p11_hash_free (configs);
 			return rv;
 		}
 	}
 
-	hash_free (configs);
+	_p11_hash_free (configs);
 	return CKR_OK;
 }
 
@@ -503,49 +483,58 @@ static CK_RV
 initialize_module_unlocked_reentrant (Module *mod)
 {
 	CK_RV rv = CKR_OK;
-
+	thread_t self;
 	assert (mod);
 
+	self = _p11_thread_self ();
+
+	if (mod->initialize_thread == self) {
+		_p11_message ("p11-kit initialization called recursively");
+		return CKR_FUNCTION_FAILED;
+	}
+
 	/*
 	 * Increase ref first, so module doesn't get freed out from
 	 * underneath us when the mutex is unlocked below.
 	 */
 	++mod->ref_count;
+	mod->initialize_thread = self;
 
-	if (!mod->initialize_count) {
+	/* Change over to the module specific mutex */
+	_p11_mutex_lock (&mod->initialize_mutex);
+	_p11_unlock ();
 
-		debug ("C_Initialize: calling");
+	if (!mod->initialize_called) {
 
-		_p11_unlock ();
+		_p11_debug ("C_Initialize: calling");
 
-			assert (mod->funcs);
-			rv = mod->funcs->C_Initialize (&mod->init_args);
+		assert (mod->funcs);
+		rv = mod->funcs->C_Initialize (&mod->init_args);
 
-		_p11_lock ();
-
-		debug ("C_Initialize: result: %lu", rv);
-
-		/*
-		 * Because we have the mutex unlocked above, two initializes could
-		 * race. Therefore we need to take CKR_CRYPTOKI_ALREADY_INITIALIZED
-		 * into account.
-		 *
-		 * We also need to take into account where in a race both calls return
-		 * CKR_OK (which is not according to the spec but may happen, I mean we
-		 * do it in this module, so it's not unimaginable).
-		 */
+		_p11_debug ("C_Initialize: result: %lu", rv);
 
+		/* Module was initialized and C_Finalize should be called */
 		if (rv == CKR_OK)
-			++mod->initialize_count;
+			mod->initialize_called = 1;
+
+		/* Module was already initialized, we don't call C_Finalize */
 		else if (rv == CKR_CRYPTOKI_ALREADY_INITIALIZED)
 			rv = CKR_OK;
-		else
-			--mod->ref_count;
 	}
 
+	_p11_mutex_unlock (&mod->initialize_mutex);
+	_p11_lock ();
+
+	/* Don't claim reference if failed */
+	if (rv != CKR_OK)
+		--mod->ref_count;
+
+	mod->initialize_thread = 0;
 	return rv;
 }
 
+#ifdef OS_UNIX
+
 static void
 reinitialize_after_fork (void)
 {
@@ -553,15 +542,15 @@ reinitialize_after_fork (void)
 	Module *mod;
 
 	/* WARNING: This function must be reentrant */
-	debug ("forked");
+	_p11_debug ("forked");
 
 	_p11_lock ();
 
 		if (gl.modules) {
-			hash_iterate (gl.modules, &iter);
-			while (hash_next (&iter, NULL, (void **)&mod)) {
-				if (mod->initialize_count > 0) {
-					mod->initialize_count = 0;
+			_p11_hash_iterate (gl.modules, &iter);
+			while (_p11_hash_next (&iter, NULL, (void **)&mod)) {
+				if (mod->initialize_called) {
+					mod->initialize_called = 0;
 
 					/* WARNING: Reentrancy can occur here */
 					initialize_module_unlocked_reentrant (mod);
@@ -574,13 +563,15 @@ reinitialize_after_fork (void)
 	_p11_kit_proxy_after_fork ();
 }
 
+#endif /* OS_UNIX */
+
 static CK_RV
 init_globals_unlocked (void)
 {
 	static int once = 0;
 
 	if (!gl.modules)
-		gl.modules = hash_create (hash_direct_hash, hash_direct_equal,
+		gl.modules = _p11_hash_create (_p11_hash_direct_hash, _p11_hash_direct_equal,
 		                          NULL, free_module_unlocked);
 	if (!gl.modules)
 		return CKR_HOST_MEMORY;
@@ -588,7 +579,9 @@ init_globals_unlocked (void)
 	if (once)
 		return CKR_OK;
 
+#ifdef OS_UNIX
 	pthread_atfork (NULL, NULL, reinitialize_after_fork);
+#endif
 	once = 1;
 
 	return CKR_OK;
@@ -601,15 +594,15 @@ free_modules_when_no_refs_unlocked (void)
 	hashiter iter;
 
 	/* Check if any modules have a ref count */
-	hash_iterate (gl.modules, &iter);
-	while (hash_next (&iter, NULL, (void **)&mod)) {
+	_p11_hash_iterate (gl.modules, &iter);
+	while (_p11_hash_next (&iter, NULL, (void **)&mod)) {
 		if (mod->ref_count)
 			return;
 	}
 
-	hash_free (gl.modules);
+	_p11_hash_free (gl.modules);
 	gl.modules = NULL;
-	hash_free (gl.config);
+	_p11_hash_free (gl.config);
 	gl.config = NULL;
 }
 
@@ -635,19 +628,20 @@ finalize_module_unlocked_reentrant (Module *mod)
 	 */
 	++mod->ref_count;
 
-	while (mod->initialize_count > 0) {
+	_p11_mutex_lock (&mod->initialize_mutex);
+	_p11_unlock ();
 
-		_p11_unlock ();
+	if (mod->initialize_called) {
 
-			assert (mod->funcs);
-			mod->funcs->C_Finalize (NULL);
+		assert (mod->funcs);
+		mod->funcs->C_Finalize (NULL);
 
-		_p11_lock ();
-
-		if (mod->initialize_count > 0)
-			--mod->initialize_count;
+		mod->initialize_called = 0;
 	}
 
+	_p11_mutex_unlock (&mod->initialize_mutex);
+	_p11_lock ();
+
 	/* Match the increment above */
 	--mod->ref_count;
 
@@ -663,8 +657,8 @@ find_module_for_name_unlocked (const char *name)
 
 	assert (name);
 
-	hash_iterate (gl.modules, &iter);
-	while (hash_next (&iter, NULL, (void **)&mod))
+	_p11_hash_iterate (gl.modules, &iter);
+	while (_p11_hash_next (&iter, NULL, (void **)&mod))
 		if (mod->ref_count && mod->name && strcmp (name, mod->name) == 0)
 			return mod;
 	return NULL;
@@ -683,8 +677,8 @@ _p11_kit_initialize_registered_unlocked_reentrant (void)
 
 	rv = load_registered_modules_unlocked ();
 	if (rv == CKR_OK) {
-		hash_iterate (gl.modules, &iter);
-		while (hash_next (&iter, NULL, (void **)&mod)) {
+		_p11_hash_iterate (gl.modules, &iter);
+		while (_p11_hash_next (&iter, NULL, (void **)&mod)) {
 
 			/* Skip all modules that aren't registered */
 			if (!mod->name)
@@ -693,7 +687,7 @@ _p11_kit_initialize_registered_unlocked_reentrant (void)
 			rv = initialize_module_unlocked_reentrant (mod);
 
 			if (rv != CKR_OK) {
-				debug ("failed to initialize module: %s: %s",
+				_p11_debug ("failed to initialize module: %s: %s",
 				       mod->name, p11_kit_strerror (rv));
 				break;
 			}
@@ -725,8 +719,10 @@ p11_kit_initialize_registered (void)
 {
 	CK_RV rv;
 
+	_p11_library_init_once ();
+
 	/* WARNING: This function must be reentrant */
-	debug ("in");
+	_p11_debug ("in");
 
 	_p11_lock ();
 
@@ -743,7 +739,7 @@ p11_kit_initialize_registered (void)
 	if (rv != CKR_OK)
 		p11_kit_finalize_registered ();
 
-	debug ("out: %lu", rv);
+	_p11_debug ("out: %lu", rv);
 	return rv;
 }
 
@@ -760,20 +756,20 @@ _p11_kit_finalize_registered_unlocked_reentrant (void)
 
 	/* WARNING: This function must be reentrant */
 
-	to_finalize = calloc (hash_size (gl.modules), sizeof (Module *));
+	to_finalize = calloc (_p11_hash_size (gl.modules), sizeof (Module *));
 	if (!to_finalize)
 		return CKR_HOST_MEMORY;
 
 	count = 0;
-	hash_iterate (gl.modules, &iter);
-	while (hash_next (&iter, NULL, (void **)&mod)) {
+	_p11_hash_iterate (gl.modules, &iter);
+	while (_p11_hash_next (&iter, NULL, (void **)&mod)) {
 
 		/* Skip all modules that aren't registered */
 		if (mod->name)
 			to_finalize[count++] = mod;
 	}
 
-	debug ("finalizing %d modules", count);
+	_p11_debug ("finalizing %d modules", count);
 
 	for (i = 0; i < count; ++i) {
 		/* WARNING: Reentrant calls can occur here */
@@ -810,8 +806,10 @@ p11_kit_finalize_registered (void)
 {
 	CK_RV rv;
 
+	_p11_library_init_once ();
+
 	/* WARNING: This function must be reentrant */
-	debug ("in");
+	_p11_debug ("in");
 
 	_p11_lock ();
 
@@ -824,7 +822,7 @@ p11_kit_finalize_registered (void)
 
 	_p11_unlock ();
 
-	debug ("out: %lu", rv);
+	_p11_debug ("out: %lu", rv);
 	return rv;
 }
 
@@ -837,10 +835,10 @@ _p11_kit_registered_modules_unlocked (void)
 	int i = 0;
 
 	if (gl.modules)
-		result = calloc (hash_size (gl.modules) + 1, sizeof (CK_FUNCTION_LIST_PTR));
+		result = calloc (_p11_hash_size (gl.modules) + 1, sizeof (CK_FUNCTION_LIST_PTR));
 	if (result) {
-		hash_iterate (gl.modules, &iter);
-		while (hash_next (&iter, NULL, (void **)&mod)) {
+		_p11_hash_iterate (gl.modules, &iter);
+		while (_p11_hash_next (&iter, NULL, (void **)&mod)) {
 			if (mod->ref_count && mod->name)
 				result[i++] = mod->funcs;
 		}
@@ -866,6 +864,8 @@ p11_kit_registered_modules (void)
 {
 	CK_FUNCTION_LIST_PTR_PTR result;
 
+	_p11_library_init_once ();
+
 	_p11_lock ();
 
 		_p11_kit_clear_message ();
@@ -896,11 +896,13 @@ p11_kit_registered_module_to_name (CK_FUNCTION_LIST_PTR module)
 	Module *mod;
 	char *name = NULL;
 
+	_p11_library_init_once ();
+
 	_p11_lock ();
 
 		_p11_kit_clear_message ();
 
-		mod = module && gl.modules ? hash_get (gl.modules, module) : NULL;
+		mod = module && gl.modules ? _p11_hash_get (gl.modules, module) : NULL;
 		if (mod && mod->name)
 			name = strdup (mod->name);
 
@@ -960,6 +962,8 @@ p11_kit_registered_option (CK_FUNCTION_LIST_PTR module, const char *field)
 	char *option = NULL;
 	hashmap *config = NULL;
 
+	_p11_library_init_once ();
+
 	_p11_lock ();
 
 		_p11_kit_clear_message ();
@@ -968,13 +972,13 @@ p11_kit_registered_option (CK_FUNCTION_LIST_PTR module, const char *field)
 			config = gl.config;
 
 		} else {
-			mod = gl.modules ? hash_get (gl.modules, module) : NULL;
+			mod = gl.modules ? _p11_hash_get (gl.modules, module) : NULL;
 			if (mod)
 				config = mod->config;
 		}
 
 		if (config && field) {
-			option = hash_get (config, field);
+			option = _p11_hash_get (config, field);
 			if (option)
 				option = strdup (option);
 		}
@@ -1021,8 +1025,10 @@ p11_kit_initialize_module (CK_FUNCTION_LIST_PTR module)
 	Module *mod;
 	CK_RV rv = CKR_OK;
 
+	_p11_library_init_once ();
+
 	/* WARNING: This function must be reentrant for the same arguments */
-	debug ("in");
+	_p11_debug ("in");
 
 	_p11_lock ();
 
@@ -1031,9 +1037,9 @@ p11_kit_initialize_module (CK_FUNCTION_LIST_PTR module)
 		rv = init_globals_unlocked ();
 		if (rv == CKR_OK) {
 
-			mod = hash_get (gl.modules, module);
+			mod = _p11_hash_get (gl.modules, module);
 			if (mod == NULL) {
-				debug ("allocating new module");
+				_p11_debug ("allocating new module");
 				allocated = mod = alloc_module_unlocked ();
 				if (mod == NULL)
 					rv = CKR_HOST_MEMORY;
@@ -1043,7 +1049,7 @@ p11_kit_initialize_module (CK_FUNCTION_LIST_PTR module)
 
 			/* If this was newly allocated, add it to the list */
 			if (rv == CKR_OK && allocated) {
-				if (hash_set (gl.modules, allocated->funcs, allocated))
+				if (_p11_hash_set (gl.modules, allocated->funcs, allocated))
 					allocated = NULL;
 				else
 					rv = CKR_HOST_MEMORY;
@@ -1070,7 +1076,7 @@ p11_kit_initialize_module (CK_FUNCTION_LIST_PTR module)
 
 	_p11_unlock ();
 
-	debug ("out: %lu", rv);
+	_p11_debug ("out: %lu", rv);
 	return rv;
 }
 
@@ -1104,16 +1110,18 @@ p11_kit_finalize_module (CK_FUNCTION_LIST_PTR module)
 	Module *mod;
 	CK_RV rv = CKR_OK;
 
+	_p11_library_init_once ();
+
 	/* WARNING: This function must be reentrant for the same arguments */
-	debug ("in");
+	_p11_debug ("in");
 
 	_p11_lock ();
 
 		_p11_kit_clear_message ();
 
-		mod = gl.modules ? hash_get (gl.modules, module) : NULL;
+		mod = gl.modules ? _p11_hash_get (gl.modules, module) : NULL;
 		if (mod == NULL) {
-			debug ("module not found");
+			_p11_debug ("module not found");
 			rv = CKR_ARGUMENTS_BAD;
 		} else {
 			/* WARNING: Rentrancy can occur here */
@@ -1124,7 +1132,7 @@ p11_kit_finalize_module (CK_FUNCTION_LIST_PTR module)
 
 	_p11_unlock ();
 
-	debug ("out: %lu", rv);
+	_p11_debug ("out: %lu", rv);
 	return rv;
 }
 
@@ -1168,8 +1176,10 @@ p11_kit_load_initialize_module (const char *module_path,
 	Module *mod;
 	CK_RV rv = CKR_OK;
 
+	_p11_library_init_once ();
+
 	/* WARNING: This function must be reentrant for the same arguments */
-	debug ("in: %s", module_path);
+	_p11_debug ("in: %s", module_path);
 
 	_p11_lock ();
 
@@ -1201,6 +1211,6 @@ p11_kit_load_initialize_module (const char *module_path,
 
 	_p11_unlock ();
 
-	debug ("out: %lu", rv);
+	_p11_debug ("out: %lu", rv);
 	return rv;
 }
