@@ -36,6 +36,7 @@
 
 #include "config.h"
 
+#define DEBUG_FLAG DEBUG_LIB
 #include "debug.h"
 #include "p11-kit.h"
 #include "private.h"
@@ -308,7 +309,8 @@ _p11_library_get_thread_local (void)
 void
 _p11_library_init (void)
 {
-	debug_init ();
+	_p11_debug_init ();
+	_p11_debug ("initializing library");
 	_p11_mutex_init (&_p11_mutex);
 	thread_local = TlsAlloc ();
 }
@@ -317,6 +319,8 @@ static void
 free_tls_value (LPVOID data)
 {
 	p11_local *local = data;
+	if (local == NULL)
+		return;
 	if (local->last_error)
 		LocalFree (local->last_error);
 	LocalFree (data);
@@ -327,6 +331,8 @@ _p11_library_uninit (void)
 {
 	LPVOID data;
 
+	_p11_debug ("uninitializing library");
+
 	if (thread_local != TLS_OUT_OF_INDEXES) {
 		data = TlsGetValue (thread_local);
 		free_tls_value (data);
@@ -345,12 +351,15 @@ DllMain (HINSTANCE instance,
 	switch (reason) {
 	case DLL_PROCESS_ATTACH:
 		_p11_library_init ();
-		if (thread_local == TLS_OUT_OF_INDEXES)
+		if (thread_local == TLS_OUT_OF_INDEXES) {
+			_p11_debug ("couldn't setup tls");
 			return FALSE;
+		}
 		break;
 
 	case DLL_THREAD_DETACH:
 		if (thread_local != TLS_OUT_OF_INDEXES) {
+			_p11_debug ("thread stopped, freeing tls");
 			data = TlsGetValue (thread_local);
 			free_tls_value (data);
 		}
